@@ -60,151 +60,11 @@ __all__ = ["Std",
            "load_model"]
 
 
-class Std(_Layer):
-    """计算每个序列各stride的标准差.
-
-    Notes:
-        计算每个feature各个stride的standard deviation
-
-    """
-
-    def __init__(self, stride: int = 10, **kwargs):
-        """标准差.
-
-        Args:
-            stride (int): time steps需要是stride的整数倍
-
-        """
-        if stride <= 1:
-            raise ValueError("Illegal Argument: stride should be "
-                             "greater than 1")
-        super(Std, self).__init__(**kwargs)
-        self.stride = stride
-        self.intermediate_shape = None
-
-    def build(self, input_shape):
-        """构建该层，计算维度信息."""
-        (features,
-         output_length) = __get_dimensions__(input_shape, self.stride)
-        self.intermediate_shape = [-1,
-                                   output_length,
-                                   self.stride,
-                                   features]
-
-    def call(self, inputs, *args, **kwargs):
-        """函数主逻辑实现部分.
-
-        Args:
-            inputs (tensor): 输入dimension为(batch_size, time_steps, features)
-
-        Returns:
-            dimension 为(batch_size, time_steps / stride, features)
-
-        """
-        # compute means for each stride
-        means = _tf.repeat(
-            _tf.nn.avg_pool(inputs,
-                            ksize=self.stride,
-                            strides=self.stride,
-                            padding="VALID"),
-            self.stride,
-            axis=1
-        )
-
-        # subtract means for each stride
-        squared_diff = _tf.square(_tf.subtract(inputs, means))
-        squared_diff = _tf.reshape(squared_diff, self.intermediate_shape)
-
-        # compute standard deviation for each stride
-        mean_squared_diff = _tf.reduce_mean(squared_diff, axis=2)
-        std = _tf.sqrt(mean_squared_diff)
-
-        return std
-
-    def get_config(self):
-        """获取参数，保存模型需要的函数."""
-        config = super().get_config().copy()
-        config.update({'stride': self.stride})
-        return config
-
-
-class ZScore(_Layer):
-    """计算每个序列各stride的均值除以其标准差.
-
-    Notes:
-        并非严格意义上的z-score,
-        计算公式为每个feature各个stride的mean除以各自的standard deviation
-
-    """
-
-    def __init__(self, stride: int = 10, **kwargs):
-        """均值除以标准差.
-
-        Args:
-            stride (int): time steps需要是stride的整数倍
-
-        """
-        if stride <= 1:
-            raise ValueError("Illegal Argument: stride should be "
-                             "greater than 1")
-        super(ZScore, self).__init__(**kwargs)
-        self.stride = stride
-        self.intermediate_shape = None
-
-    def build(self, input_shape):
-        """构建该层，计算维度信息."""
-        (features,
-         output_length) = __get_dimensions__(input_shape, self.stride)
-        self.intermediate_shape = [-1,
-                                   output_length,
-                                   self.stride,
-                                   features]
-
-    def call(self, inputs, *args, **kwargs):
-        """函数主逻辑实现部分.
-
-        Args:
-            inputs (tensor): 输入dimension为(batch_size, time_steps, features)
-
-        Returns:
-            dimension 为(batch_size, time_steps / stride, features)
-
-        """
-        # compute means for each stride
-        means = _tf.nn.avg_pool(inputs,
-                                ksize=self.stride,
-                                strides=self.stride,
-                                padding="VALID")
-
-        # compute standard deviations for each stride
-        means_broadcast = _tf.repeat(means, self.stride, axis=1)
-        squared_diff = _tf.square(_tf.subtract(inputs, means_broadcast))
-        squared_diff = _tf.reshape(squared_diff, self.intermediate_shape)
-        mean_squared_diff = _tf.reduce_mean(squared_diff, axis=2)
-        std = _tf.sqrt(mean_squared_diff)
-
-        # divide means by standard deviations for each stride
-        z_score = _tf.math.divide_no_nan(means, std)
-        return z_score
-
-    def get_config(self):
-        """获取参数，保存模型需要的函数."""
-        config = super().get_config().copy()
-        config.update({'stride': self.stride})
-        return config
-
-
-class LinearDecay(_Layer):
-    """计算每个序列各stride的线性衰减加权平均.
-
-    Notes:
-        以线性衰减为权重，计算每个feature各个stride的均值：
-        如stride为10，则某feature该stride的权重为(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
-
-    """
+class _StrideLayer(_Layer, _ABC):
+    """计算每个stride的统计值的基类."""
 
     def __init__(self, stride=10, **kwargs):
-        """线性递减加权平均.
+        """计算每个stride的统计值的基类.
 
         Args:
             stride (int): time steps需要是stride的整数倍
@@ -213,7 +73,7 @@ class LinearDecay(_Layer):
         if stride <= 1:
             raise ValueError("Illegal Argument: stride should be "
                              "greater than 1")
-        super(LinearDecay, self).__init__(**kwargs)
+        super(_StrideLayer, self).__init__(**kwargs)
         self.stride = stride
         self.out_shape = None
         self.intermediate_shape = None
@@ -225,6 +85,79 @@ class LinearDecay(_Layer):
         self.out_shape = [-1, output_length, features]
         self.intermediate_shape = [-1, self.stride, features]
 
+    def get_config(self):
+        """获取参数，保存模型需要的函数."""
+        config = super().get_config().copy()
+        config.update({'stride': self.stride})
+        return config
+
+
+class Std(_StrideLayer):
+    """计算每个序列各stride的标准差.
+
+    Notes:
+        计算每个feature各个stride的standard deviation
+
+    """
+
+    def call(self, inputs, *args, **kwargs):
+        """函数主逻辑实现部分.
+
+        Args:
+            inputs (tensor): 输入dimension为(batch_size, time_steps, features)
+
+        Returns:
+            dimension 为(batch_size, time_steps / stride, features)
+
+        """
+        strides = _tf.reshape(inputs, self.intermediate_shape)
+
+        # compute standard deviations for each stride
+        std = _tf.math.reduce_std(strides, axis=-2)
+        return _tf.reshape(std, self.out_shape)
+
+
+class ZScore(_StrideLayer):
+    """计算每个序列各stride的均值除以其标准差.
+
+    Notes:
+        并非严格意义上的z-score,
+        计算公式为每个feature各个stride的mean除以各自的standard deviation
+
+    """
+
+    def call(self, inputs, *args, **kwargs):
+        """函数主逻辑实现部分.
+
+        Args:
+            inputs (tensor): 输入dimension为(batch_size, time_steps, features)
+
+        Returns:
+            dimension 为(batch_size, time_steps / stride, features)
+
+        """
+        strides = _tf.reshape(inputs, self.intermediate_shape)
+
+        # compute standard deviations for each stride
+        std = _tf.math.reduce_std(strides, axis=-2)
+
+        # compute means for each stride
+        means = _tf.math.reduce_mean(strides, axis=-2)
+
+        # divide means by standard deviations for each stride
+        z_score = _tf.math.divide_no_nan(means, std)
+        return _tf.reshape(z_score, self.out_shape)
+
+
+class LinearDecay(_StrideLayer):
+    """计算每个序列各stride的线性衰减加权平均.
+
+    Notes:
+        以线性衰减为权重，计算每个feature各个stride的均值：
+        如stride为10，则某feature该stride的权重为(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
+
+    """
+
     def call(self, inputs, *args, **kwargs):
         """函数主逻辑实现部分.
 
@@ -250,12 +183,6 @@ class LinearDecay(_Layer):
         linear_decay = _tf.reshape(linear_decay, self.out_shape)
         return linear_decay
 
-    def get_config(self):
-        """获取参数，保存模型需要的函数."""
-        config = super().get_config().copy()
-        config.update({'stride': self.stride})
-        return config
-
 
 class Return(_Layer):
     """计算每个序列各stride的回报率.
@@ -634,6 +561,7 @@ class AlphaNetV3(_Model):
                  classification=False,
                  categories=0,
                  recurrent_unit="GRU",
+                 hidden_units=30,
                  *args,
                  **kwargs):
         """Alpha net v3.
@@ -656,13 +584,13 @@ class AlphaNetV3(_Model):
         self.expanded5 = FeatureExpansion(stride=5)
         self.normalized10 = _tfl.BatchNormalization()
         self.normalized5 = _tfl.BatchNormalization()
-        self.dropout = _tfl.Dropout(self.dropout)
+        self.dropout_layer = _tfl.Dropout(self.dropout)
         if recurrent_unit == "GRU":
-            self.recurrent10 = _tfl.GRU(units=30)
-            self.recurrent5 = _tfl.GRU(units=30)
+            self.recurrent10 = _tfl.GRU(units=hidden_units)
+            self.recurrent5 = _tfl.GRU(units=hidden_units)
         elif recurrent_unit == "LSTM":
-            self.recurrent10 = _tfl.LSTM(units=30)
-            self.recurrent5 = _tfl.LSTM(units=30)
+            self.recurrent10 = _tfl.LSTM(units=hidden_units)
+            self.recurrent5 = _tfl.LSTM(units=hidden_units)
         else:
             raise ValueError("Unknown recurrent_unit")
         self.normalized10_2 = _tfl.BatchNormalization()
@@ -697,7 +625,7 @@ class AlphaNetV3(_Model):
         normalized10_2 = self.normalized10_2(recurrent10, training=training)
         normalized5_2 = self.normalized5_2(recurrent5, training=training)
         concat = self.concat([normalized10_2, normalized5_2])
-        dropout = self.dropout(concat, training=training)
+        dropout = self.dropout_layer(concat, training=training)
         output = self.outputs(dropout)
         return output
 
@@ -739,10 +667,16 @@ class AlphaNetV4(_Model):
                  classification=False,
                  categories=0,
                  recurrent_unit="GRU",
+                 hidden_units=30,
                  *args,
                  **kwargs):
         """Alpha net v4.
 
+        Notes:
+            去掉了batch normalization的模型，
+            训练需要使用data模块的normalization
+            或其他自定义normalization.
+
         Args:
             dropout: 跟在特征扩张以及Batch Normalization之后的dropout，默认无dropout
             l2: 输出层的l2-regularization参数
@@ -756,25 +690,17 @@ class AlphaNetV4(_Model):
         self.dropout = dropout
         self.expanded10 = FeatureExpansion(stride=10)
         self.expanded5 = FeatureExpansion(stride=5)
-        self.normalized10 = _tfl.BatchNormalization()
-        self.normalized5 = _tfl.BatchNormalization()
-        self.dropout = _tfl.Dropout(self.dropout)
+        self.dropout_layer = _tfl.Dropout(self.dropout)
         if recurrent_unit == "GRU":
-            self.recurrent10 = _tfl.GRU(units=30)
-            self.recurrent5 = _tfl.GRU(units=30)
+            self.recurrent10 = _tfl.GRU(units=hidden_units)
+            self.recurrent5 = _tfl.GRU(units=hidden_units)
         elif recurrent_unit == "LSTM":
-            self.recurrent10 = _tfl.LSTM(units=30)
-            self.recurrent5 = _tfl.LSTM(units=30)
+            self.recurrent10 = _tfl.LSTM(units=hidden_units)
+            self.recurrent5 = _tfl.LSTM(units=hidden_units)
         else:
             raise ValueError("Unknown recurrent_unit")
-        self.normalized10_2 = _tfl.BatchNormalization()
-        self.normalized5_2 = _tfl.BatchNormalization()
         self.concat = _tfl.Concatenate(axis=-1)
         self.regularizer = _tf.keras.regularizers.l2(self.l2)
-        self.dense = _tfl.Dense(units=30,
-                                activation="relu",
-                                kernel_initializer="he_normal",
-                                kernel_regularizer=self.regularizer)
         if classification:
             if categories < 1:
                 raise ValueError("categories should be at least 1")
@@ -796,16 +722,11 @@ class AlphaNetV4(_Model):
         """计算逻辑实现."""
         expanded10 = self.expanded10(inputs)
         expanded5 = self.expanded5(inputs)
-        normalized10 = self.normalized10(expanded10, training=training)
-        normalized5 = self.normalized5(expanded5, training=training)
-        recurrent10 = self.recurrent10(normalized10)
-        recurrent5 = self.recurrent5(normalized5)
-        normalized10_2 = self.normalized10_2(recurrent10, training=training)
-        normalized5_2 = self.normalized5_2(recurrent5, training=training)
-        concat = self.concat([normalized10_2, normalized5_2])
-        dropout = self.dropout(concat, training=training)
-        dense = self.dense(dropout)
-        output = self.outputs(dense)
+        recurrent10 = self.recurrent10(expanded10)
+        recurrent5 = self.recurrent5(expanded5)
+        concat = self.concat([recurrent10, recurrent5])
+        dropout = self.dropout_layer(concat, training=training)
+        output = self.outputs(dropout)
         return output
 
     def compile(self,
